require 'json'
require 'uu/os/lang/json_patch'
require 'uu/os/lang/hash_converter'
require 'uu/os/http/http_request'
require 'uu/os/http/multipart_parser'

module UU
  module OS
    module Server

      # Module providing method for parsing of request parameters.
      module ParameterParser

        # Parses request parameters. Parameters are returned as single +Hash+ where
        # query parameters are merged with parameters in body (query parameters are
        # overwritten in case same parameter is contained in body).
        # @param [UU::OS::HTTP::HTTPRequest] request Instance of request
        # @param [Symbol, String] binary_name Name of attribute containing large
        #    binary data. In case attribute with this name is found in body of multipart
        #    request, its content is returned as direct (unbuffered) stream and rest of
        #    multipart body is dumped (binary attribute should be always the last item of
        #    multipart body).
        # @return [Hash<Symbol, Object>] Parsed request parameters.
        def parse_parameters(request, binary_name = nil)
          ParameterParser.parse_parameters(request, binary_name)
        end

        # @see #find_classes
        # @private
        def self.parse_parameters(request, binary_name = nil) # :nodoc:
          parameters_class = Class.new(Hash)
          parameters_class.send(:include, UU::OS::Lang::HashConverter)
          parameters = parameters_class.new

          # Read query parameters and try to deserialize nested JSON
          parameters.merge!(request.parameters)
          parameters = parameters.inject({}) do |tmp, (key, value)|
            if value =~ /(^{.*}$)|(^\[.*\]$)/
              tmp[key] = (JSON.parse(value, symbolize_names: true) rescue value)
            else
              tmp[key] = value
            end
            tmp
          end

          # Read parameters from body which can be either sent as multipart/form-data...
          if request.headers.has_key?(:content_type) && request.headers[:content_type][0] =~ /multipart\/form-data;.*boundary=([^;]+)/
            UU::OS::HTTP::MultipartParser.parse_multipart(request.body, $1, binary_name).each do |item|
              name = item[:name].to_sym
              body = item[:body]
              if item[:content_type] =~ /application\/json/ && body.kind_of?(String)
                parameters[name] = (JSON.parse(body, symbolize_names: true, quirks_mode: true) rescue body)
              else
                parameters[name] = body
              end
              if binary_name =~ /^#{name}$/
                parameters[:name] ||= item[:filename]
                parameters[:contentType] ||= item[:content_type]
              end
            end
          else
            # ... or must be valid JSON containing Hash
            body = request.body
            body = body.read if body.respond_to?:read
            if !body.nil? && !body.empty?
              begin
                body = JSON.parse(body, symbolize_names: true, quirks_mode: true)
              rescue
                raise TypeError, 'content is not valid JSON'
              end
              if !body.kind_of?(Hash)
                raise TypeError, 'JSON does not represent key-value structure'
              end
              parameters.merge!(body)
            end
          end
          return parameters
        end

      end

    end
  end
end
