require 'rack/utils'
require 'uu/os/http/http_request'

module UU
  module OS
    module Server

      # Class used for dynamic invocation of servlets so they
      # need not to implement Rack API.
      # @private
      class RackDispatcher # :nodoc:

        # Creates new instance of dispatcher.
        # @param [Rack::Builder] builder Instance of rack builder
        # @param [Class<UU::OS::Server::AbstractServlet>] servlet Servlet class.
        # @private
        def initialize(builder, servlet) # :nodoc:
          @servlet = servlet.new
        end

        # Main service method invoked by Rack framework.
        # @private
        def call(env) # :nodoc:
          request = read_request(env)
          response = @servlet.service(request)
          write_response(response)
        end

        protected

        # Transforms native Rack request into abstract request object.
        # @private
        def read_request(env) # :nodoc:
          url = env['REQUEST_URI'].to_s.split('?')[0]
          method = env['REQUEST_METHOD']
          headers = env.inject({}) do |tmp, (key, value)|
            if key =~ /^HTTP_/
              tmp[key[5..-1]] = value.split(',').map(&:strip)
            elsif key =~ /^CONTENT/
              tmp[key] = value.split(',').map(&:strip)
            end
            tmp
          end
          params = Rack::Utils.parse_nested_query(env['QUERY_STRING'].to_s, '&')
          # Actual support of streaming depends on particular implementation of
          # rack handler. According to rack specification "rack.input" must be
          # rewindable (and thus buffered in memory or tempfile), however some
          # servers allow to override this behavior.
          body = env['rack.input']
          UU::OS::HTTP::HTTPRequest.new(url, method, headers, params, body)
        end

        # Writes abstract response to Rack response.
        # @private
        def write_response(response_data = {}) # :nodoc:
          body = response_data[:body] || ''
          if body.respond_to?(:read)
            body = ChunkedIO.new(body)
          elsif !body.respond_to?(:each)
            body = [body]
          end
          # Actual support of streaming depends on particular implementation of
          # rack handler. Body is read using each, however handler implementation
          # provide output into which body is written.
          [response_data[:status], response_data[:headers], body]
        end

        # Class allowing chunked streaming of IO response body (by default
        # each on IO is in fact "each_line" and therefore is not suitable
        # for binary data streaming).
        # @private
        class ChunkedIO # :nodoc:
          def initialize(io)
            @io = io
          end

          def each(&block)
            while (!(chunk = @io.read(65535)).nil?)
              block.yield(chunk)
            end
            @io.close if @io.respond_to?(:close)
          end
        end

      end

    end
  end
end
